# -*- coding: utf-8 -*-
"""
This module provides testing code for the
auxi.modeling.business.entity module.

@author: Ex Mente Technologies (Pty) Ltd
"""

import unittest
import collections
from datetime import datetime
from dateutil.relativedelta import relativedelta
from auxi.modeling.business.clock import Clock
from auxi.modeling.business.entity import Entity
from auxi.modeling.business.basicactivity import BasicActivity
from auxi.modeling.financial.des.generalledgeraccount import AccountType
from auxi.modeling.financial.des.generalledgerstructure import GeneralLedgerStructure
from auxi.modeling.financial.des.transactiontemplate import TransactionTemplate

__version__ = "0.2.0"


# =============================================================================
# Types.
# =============================================================================

class TestAllFunctions(unittest.TestCase):
    """
      Tester for the auxi.modeling.entity.component class.
    """

    def setUp(self):
        self.gl_structure = GeneralLedgerStructure(
            "NameA",
            description="DescriptionA")
        self.reva_acc = self.gl_structure.create_account(
            "RevA",
            description="RevA_Desc",
            number="011",
            account_type=AccountType.revenue)
        self.expa_acc = self.gl_structure.create_account(
            "ExpenseA",
            description="ExpenseA_Desc",
            number="011",
            account_type=AccountType.expense)

        self.object = Entity("EntityA",
                             gl_structure=self.gl_structure,
                             description="DescriptionA")
        # Set up the needed objects
        comp1 = self.object.create_component("ComponentA1", description="ca1")
        self.tx_template = TransactionTemplate("NameA",
                                               description="DescriptionA",
                                               dt_account="Bank",
                                               cr_account="Sales")
        basic_activity = BasicActivity("BasicActivityA",
                                       description="DescriptionA",
                                       start=datetime(2016, 2, 1),
                                       end=datetime(2017, 2, 1),
                                       interval=1,
                                       amount=5000,
                                       tx_template=self.tx_template)
        comp1.activities.append(basic_activity)

        self.clock = Clock("NameA", start_datetime=datetime(2016, 1, 1))

    def test_constructor(self):
        self.assertEqual(self.object.name, "EntityA")
        self.assertEqual(self.object.description, "DescriptionA")
        self.assertEqual(self.object.gl.structure.name, self.gl_structure.name)
        self.assertEqual(len(self.object.tax_rule_set.rules), 0)
        self.assertEqual(self.object.period_count, -1)

    def test_set_parent_path(self):
        self.object.set_parent_path("modelA")
        self.assertEqual(self.object.path, "modelA/EntityA")

    def test_set_name(self):
        self.object.set_parent_path("modelA")
        self.object.name = "NameAt"
        self.assertEqual(self.object.name, "NameAt")
        self.assertEqual(self.object.components[0].path,
                         "modelA/NameAt/ComponentA1")
        self.assertEqual(self.object.components[0].activities[0].path,
                         "modelA/NameAt/ComponentA1/BasicActivityA")

    def test_create_component(self):
        new_comp = self.object.create_component("ComponentA2",
                                                description="ca2")
        self.assertEqual(new_comp.name, "ComponentA2")
        self.assertEqual(new_comp.description, "ca2")

        self.assertEqual(new_comp, self.object.components[1])

    def test_remove_component(self):
        self.object.create_component("ComponentA2",
                                     description="ca2")
        self.object.remove_component("ComponentA2")
        self.assertEqual(len(self.object.components), 1)

    def test_prepare_to_run(self):
        # Test the prepare to run method before any run has been performed
        self.object.gl.create_transaction(
            "TestA",
            description="TestA_Desc",
            tx_datetime=datetime(2016, 2, 1),
            dt_account="Bank",
            cr_account="Sales",
            source="Peanut Sales",
            amount=20.00)
        self.object.negative_income_tax_total = 99

        self.object.prepare_to_run(self.clock, 18)

        # Was the component's prepare to run executed?
        self.assertEqual(
            self.object.components[0].activities[0].start_period_ix, 2)
        self.assertEqual(
            self.object.components[0].activities[0].end_period_ix, 14)
        # Was the datetime's set correctly?
        self.assertEqual(
            self.object._exec_year_end_datetime,
            self.clock.get_datetime_at_period_ix(18))
        self.assertEqual(
            self.object._prev_year_end_datetime,
            self.clock.start_datetime)
        self.assertEqual(
            self.object._curr_year_end_datetime,
            self.clock.start_datetime + relativedelta(years=1))
        # Was the transactions cleared?
        self.assertEqual(len(self.object.gl.transactions), 0)
        # Was the negative tax income reset to 0?
        self.assertEqual(self.object.negative_income_tax_total, 0)

    def test_run(self):
        self.object.prepare_to_run(self.clock, 20)
        self.clock.tick()
        self.clock.tick()
        self.object.run(self.clock)
        self.assertEqual(len(self.object.gl.transactions), 1)
        self.assertEqual(self.object.gl.transactions[0].amount, 5000)

    def test__perform_year_end_gross_profit_and_income_summary(self):
        gls = self.object.gl.structure
        year_end_datetime = datetime(2017, 2, 1)
        # Create the accounts to write of dict
        income_summary_write_off_accounts = collections.OrderedDict([
            (gls.costofsales_account, 700),
            (gls.incometaxexpense_account, -300),
            (gls.sales_account, 70),
            (gls.accounts[0], -30)])
        gross_profit_write_off_accounts = collections.OrderedDict([
            (gls.sales_account, 70),
            (gls.accounts[0], -30)])
        # Test when there the gross profit is 0.
        # The gross profit account should not be settled:
        inc = self.object._perform_year_end_gross_profit_and_income_summary(
            year_end_datetime,
            {},
            income_summary_write_off_accounts)
        # The gross profit account should not be settled.
        # Only the 4 transactions from the income summary year end calcs
        # should be there
        self.assertEqual(len(self.object.gl.transactions), 4)
        # The income summary should be that of the income summary year end calc
        # method: -360 (See the income summary year end calc, in this case,
        # gross profit starts at 0 instead of 5000, thus -360 instead of 4640)
        self.assertEqual(inc, -360)

        # Test when there is gross profit.
        # The gross profit account should not be settled:
        del self.object.gl.transactions[:]  # Clear the tx list
        self.object._perform_year_end_gross_profit_and_income_summary(
            year_end_datetime,
            gross_profit_write_off_accounts, {})
        # The gross profit account should be settled.
        # Only the 2 transactions from the income summary year end calcs
        # should be there as well as the gross profit settle transaction
        self.assertEqual(len(self.object.gl.transactions), 3)
        # Make sure the gross profit settle transaction appears correct
        # When gross profit is positive:
        self.assertEqual(self.object.gl.transactions[2].dt_account,
                         gls.gross_profit_account.name)
        self.assertEqual(self.object.gl.transactions[2].cr_account,
                         gls.incomesummary_account.name)
        self.assertEqual(self.object.gl.transactions[2].tx_datetime,
                         year_end_datetime)
        self.assertEqual(self.object.gl.transactions[2].amount, 40)
        self.assertEqual(self.object.gl.transactions[2].is_closing_cr_account,
                         True)
        # When gross profit is negative:
        gross_profit_write_off_accounts = collections.OrderedDict([
            (gls.sales_account, -70),
            (gls.accounts[0], 30)])
        del self.object.gl.transactions[:]  # Clear the tx list
        self.object._perform_year_end_gross_profit_and_income_summary(
            year_end_datetime,
            gross_profit_write_off_accounts, {})
        self.assertEqual(self.object.gl.transactions[2].dt_account,
                         gls.incomesummary_account.name)
        self.assertEqual(self.object.gl.transactions[2].cr_account,
                         gls.gross_profit_account.name)
        self.assertEqual(self.object.gl.transactions[2].tx_datetime,
                         year_end_datetime)
        self.assertEqual(self.object.gl.transactions[2].amount, 40)
        self.assertEqual(self.object.gl.transactions[2].is_closing_cr_account,
                         True)


    def test__perform_year_end_gross_profit(self):
        gls = self.object.gl.structure
        year_end_datetime = datetime(2017, 2, 1)

        # Create the accounts to write of dict
        gross_profit_write_off_accounts = collections.OrderedDict([
            (gls.sales_account, 70),
            (gls.accounts[0], -30)])
        gross_profit = self.object._perform_year_end_gross_profit(
            year_end_datetime, gross_profit_write_off_accounts)
        self.assertEqual(len(self.object.gl.transactions), 2)
        # Expenses: 700-300 = 400
        # Other: 70-30 = 40
        # Total: gross_profit + Other - expenses: 5000 + 40 - 400 = 4640
        self.assertEqual(gross_profit, 40)

        # Test the "gls.sales_account: 70" transaction
        self.assertEqual(self.object.gl.transactions[0].dt_account,
                         gls.gross_profit_account.name)
        self.assertEqual(self.object.gl.transactions[0].cr_account,
                         gls.sales_account.name)
        self.assertEqual(self.object.gl.transactions[0].tx_datetime,
                         year_end_datetime)
        self.assertEqual(self.object.gl.transactions[0].amount, 70)
        self.assertEqual(self.object.gl.transactions[0].is_closing_cr_account,
                         True)

        # Test the "gls.accounts[0]: -30" transaction
        self.assertEqual(self.object.gl.transactions[1].dt_account,
                         gls.accounts[0].name)
        self.assertEqual(self.object.gl.transactions[1].cr_account,
                         gls.gross_profit_account.name)
        self.assertEqual(self.object.gl.transactions[1].tx_datetime,
                         year_end_datetime)
        self.assertEqual(self.object.gl.transactions[1].amount, 30)
        self.assertEqual(self.object.gl.transactions[1].is_closing_cr_account,
                         True)

    def test__perform_year_end_income_summary(self):
        gls = self.object.gl.structure
        year_end_datetime = datetime(2017, 2, 1)
        gross_profit = 5000

        # Create the accounts to write of dict
        income_summary_write_off_accounts = collections.OrderedDict([
            (gls.costofsales_account, 700),
            (gls.incometaxexpense_account, -300),
            (gls.sales_account, 70),
            (gls.accounts[0], -30)])
        income_summary_amount = self.object._perform_year_end_income_summary(
            year_end_datetime, gross_profit, income_summary_write_off_accounts)
        self.assertEqual(len(self.object.gl.transactions), 4)
        # Expenses: 700-300 = 400
        # Other: 70-30 = 40
        # Total: gross_profit + Other - expenses: 5000 + 40 - 400 = 4640
        self.assertEqual(income_summary_amount, 4640)

        # Test the "gls.costofsales_account: 700" transaction
        self.assertEqual(self.object.gl.transactions[0].dt_account,
                         gls.incomesummary_account.name)
        self.assertEqual(self.object.gl.transactions[0].cr_account,
                         gls.costofsales_account.name)
        self.assertEqual(self.object.gl.transactions[0].tx_datetime,
                         year_end_datetime)
        self.assertEqual(self.object.gl.transactions[0].amount, 700)
        self.assertEqual(self.object.gl.transactions[0].is_closing_cr_account,
                         True)

        # Test the "gls.incometaxexpense_account: -300" transaction
        self.assertEqual(self.object.gl.transactions[1].dt_account,
                         gls.incometaxexpense_account.name)
        self.assertEqual(self.object.gl.transactions[1].cr_account,
                         gls.incomesummary_account.name)
        self.assertEqual(self.object.gl.transactions[1].tx_datetime,
                         year_end_datetime)
        self.assertEqual(self.object.gl.transactions[1].amount, 300)
        self.assertEqual(self.object.gl.transactions[1].is_closing_cr_account,
                         True)

        # Test the "gls.sales_account: 70" transaction
        self.assertEqual(self.object.gl.transactions[2].dt_account,
                         gls.sales_account.name)
        self.assertEqual(self.object.gl.transactions[2].cr_account,
                         gls.incomesummary_account.name)
        self.assertEqual(self.object.gl.transactions[2].tx_datetime,
                         year_end_datetime)
        self.assertEqual(self.object.gl.transactions[2].amount, 70)
        self.assertEqual(self.object.gl.transactions[2].is_closing_cr_account,
                         True)

        # Test the "gls.accounts[0]: -30" transaction
        self.assertEqual(self.object.gl.transactions[3].dt_account,
                         gls.incomesummary_account.name)
        self.assertEqual(self.object.gl.transactions[3].cr_account,
                         gls.accounts[0].name)
        self.assertEqual(self.object.gl.transactions[3].tx_datetime,
                         year_end_datetime)
        self.assertEqual(self.object.gl.transactions[3].amount, 30)
        self.assertEqual(self.object.gl.transactions[3].is_closing_cr_account,
                         True)

    def test__perform_year_end_income_tax(self):
        # TODO
        pass

    def test__perform_year_end_retained_earnings(self):
        gls = self.object.gl.structure
        year_end_datetime = datetime(2017, 2, 1)
        # Test for == 0 case
        income_summary = 0
        self.object._perform_year_end_retained_earnings(
            year_end_datetime, income_summary)
        self.assertEqual(len(self.object.gl.transactions), 0)

        # Test for > 0 case
        income_summary = 1000
        self.object._perform_year_end_retained_earnings(
            year_end_datetime, income_summary)
        self.assertEqual(len(self.object.gl.transactions), 1)
        self.assertEqual(self.object.gl.transactions[0].dt_account,
                         gls.incomesummary_account.name)
        self.assertEqual(self.object.gl.transactions[0].cr_account,
                         gls.retainedearnings_account.name)
        self.assertEqual(self.object.gl.transactions[0].tx_datetime,
                         year_end_datetime)
        self.assertEqual(self.object.gl.transactions[0].amount,
                         income_summary)
        self.assertEqual(self.object.gl.transactions[0].is_closing_cr_account,
                         True)

        # Test for < 0 case
        income_summary = -300
        self.object._perform_year_end_retained_earnings(
            year_end_datetime, income_summary)
        self.assertEqual(len(self.object.gl.transactions), 2)
        self.assertEqual(self.object.gl.transactions[1].dt_account,
                         gls.retainedearnings_account.name)
        self.assertEqual(self.object.gl.transactions[1].cr_account,
                         gls.incomesummary_account.name)
        self.assertEqual(self.object.gl.transactions[1].tx_datetime,
                         year_end_datetime)
        self.assertEqual(self.object.gl.transactions[1].amount,
                         income_summary)
        self.assertEqual(self.object.gl.transactions[1].is_closing_cr_account,
                         True)

    def test__perform_year_end_procedure(self):
        gls = self.object.gl.structure
        # Test when that no year end transactions are created in the
        # middle of a year
        self.object.prepare_to_run(self.clock, 28)
        self.clock.timestep_ix = 6
        # create the transactions
        self.object.gl.create_transaction(
            "Sales_tx",
            description="Sales tx",
            tx_datetime=self.clock.get_datetime_at_period_ix(
                self.clock.timestep_ix),
            dt_account=gls.accounts[0].name,  # Bank account
            cr_account=gls.sales_account.name,
            source=self.object.path,
            amount=70)
        # create the transactions
        self.object.gl.create_transaction(
            "Cost_of_Sales_tx",
            description="Cost of Sales tx",
            tx_datetime=self.clock.get_datetime_at_period_ix(
                self.clock.timestep_ix),
            dt_account=gls.costofsales_account.name,
            cr_account=gls.accounts[0].name,  # Bank account
            source=self.object.path,
            amount=30)
        self.object._perform_year_end_procedure(self.clock)
        # Only one transaction should have been created and no
        # year end transactions.
        self.assertEqual(len(self.object.gl.transactions), 2)

        # Test that the year end procedure, at the year end, creates
        #   - two gross profit transactions with values of 70 and 40
        #   - one gross profit and income summary settle transaction
        #      with a value of 70 - 30
        #   - one retained earnings transaction with a value of 70 - 30
        # Only the 2 transaction (sales tx, and cost of sales tx) and the
        # 4 'year end' transactions should have been created.)
        self.clock.timestep_ix = 12
        self.object._perform_year_end_procedure(self.clock)

        self.assertEqual(len(self.object.gl.transactions), 6)
        self.assertEqual(self.object.gl.transactions[2].amount, 70)
        self.assertEqual(self.object.gl.transactions[2].is_closing_cr_account,
                         True)
        self.assertEqual(self.object.gl.transactions[2].dt_account,
                         gls.gross_profit_account.name)
        self.assertEqual(self.object.gl.transactions[2].cr_account,
                         gls.sales_account.name)

        self.assertEqual(self.object.gl.transactions[3].amount, 30)
        self.assertEqual(self.object.gl.transactions[3].is_closing_cr_account,
                         True)
        self.assertEqual(self.object.gl.transactions[3].dt_account,
                         gls.costofsales_account.name)
        self.assertEqual(self.object.gl.transactions[3].cr_account,
                         gls.gross_profit_account.name)

        self.assertEqual(self.object.gl.transactions[4].amount, 40)
        self.assertEqual(self.object.gl.transactions[4].is_closing_cr_account,
                         True)
        self.assertEqual(self.object.gl.transactions[4].dt_account,
                         gls.gross_profit_account.name)
        self.assertEqual(self.object.gl.transactions[4].cr_account,
                         gls.incomesummary_account.name)

        self.assertEqual(self.object.gl.transactions[5].amount, 40)
        self.assertEqual(self.object.gl.transactions[5].is_closing_cr_account,
                         True)
        self.assertEqual(self.object.gl.transactions[5].dt_account,
                         gls.incomesummary_account.name)
        self.assertEqual(self.object.gl.transactions[5].cr_account,
                         gls.retainedearnings_account.name)

        # Test for a new year
        self.clock.timestep_ix = 18
        # Test for when other revenue and expense accounts transactions exists as well.

        # create a sales transactions
        self.object.gl.create_transaction(
            "SalesTx2",
            description="Sales tx2",
            tx_datetime=self.clock.get_datetime_at_period_ix(
                self.clock.timestep_ix),
            dt_account=gls.accounts[0].name,  # Bank account
            cr_account=gls.sales_account.name,
            source=self.object.path,
            amount=45)
        self.object.gl.create_transaction(
            "SalesTx3",
            description="Sales tx3",
            tx_datetime=self.clock.get_datetime_at_period_ix(
                self.clock.timestep_ix),
            dt_account=gls.accounts[0].name,  # Bank account
            cr_account=gls.sales_account.name,
            source=self.object.path,
            amount=35)
        self.object.gl.create_transaction(
            "Cost_of_Sales_tx2",
            description="Cost of Sales tx 2",
            tx_datetime=self.clock.get_datetime_at_period_ix(
                self.clock.timestep_ix),
            dt_account=gls.costofsales_account.name,
            cr_account=gls.accounts[0].name,  # Bank account
            source=self.object.path,
            amount=15)
        self.object.gl.create_transaction(
            "Cost_of_Sales_tx3",
            description="Cost of Sales tx 3",
            tx_datetime=self.clock.get_datetime_at_period_ix(
                self.clock.timestep_ix),
            dt_account=gls.costofsales_account.name,
            cr_account=gls.accounts[0].name,  # Bank account
            source=self.object.path,
            amount=5)
        # create the transactions
        self.object.gl.create_transaction(
            "revenueATx",
            description="revenue a tx",
            tx_datetime=self.clock.get_datetime_at_period_ix(
                self.clock.timestep_ix),
            dt_account=gls.accounts[0].name,  # Bank account
            cr_account=self.reva_acc.name,
            source=self.object.path,
            amount=450)
        self.object.gl.create_transaction(
            "revenueATx2",
            description="revenue a tx2",
            tx_datetime=self.clock.get_datetime_at_period_ix(
                self.clock.timestep_ix),
            dt_account=gls.accounts[0].name,  # Bank account
            cr_account=self.reva_acc.name,
            source=self.object.path,
            amount=350)
        # create the transactions
        self.object.gl.create_transaction(
            "expenseAtx",
            description="expense a tx",
            tx_datetime=self.clock.get_datetime_at_period_ix(
                self.clock.timestep_ix),
            dt_account=self.expa_acc.name,
            cr_account=gls.accounts[0].name,  # Bank account
            source=self.object.path,
            amount=320)
        self.object.gl.create_transaction(
            "expenseAtx2",
            description="expense a tx2",
            tx_datetime=self.clock.get_datetime_at_period_ix(
                self.clock.timestep_ix),
            dt_account=self.expa_acc.name,
            cr_account=gls.accounts[0].name,  # Bank account
            source=self.object.path,
            amount=180)

        # Test that the year end procedure, at the year end, creates
        #   - two gross profit transactions with values of 80 and 20
        #   - two income summary transactions with values of 800 and 500
        #   - one gross profit and income summary settle transaction
        #      with a value of 80 - 20
        #   - one retained earnings transaction with a value of
        #      800 - 500 + 80 - 20
        # Only the 6 transactions (2 sales tx, 2 cost of sales,
        # 2 revenue a and 2 expense a) and the 5 'year end' transactions
        # should have been created.)
        self.clock.timestep_ix = 24
        self.object._perform_year_end_procedure(self.clock)

        self.assertEqual(len(self.object.gl.transactions), 20)
        self.assertEqual(self.object.gl.transactions[14].amount, 80)
        self.assertEqual(self.object.gl.transactions[14].is_closing_cr_account,
                         True)
        self.assertEqual(self.object.gl.transactions[14].dt_account,
                         gls.gross_profit_account.name)
        self.assertEqual(self.object.gl.transactions[14].cr_account,
                         gls.sales_account.name)

        self.assertEqual(self.object.gl.transactions[15].amount, 20)
        self.assertEqual(self.object.gl.transactions[15].is_closing_cr_account,
                         True)
        self.assertEqual(self.object.gl.transactions[15].dt_account,
                         gls.costofsales_account.name)
        self.assertEqual(self.object.gl.transactions[15].cr_account,
                         gls.gross_profit_account.name)

        self.assertEqual(self.object.gl.transactions[16].amount, 800)
        self.assertEqual(self.object.gl.transactions[16].is_closing_cr_account,
                         True)
        self.assertEqual(self.object.gl.transactions[16].dt_account,
                         self.reva_acc.name)
        self.assertEqual(self.object.gl.transactions[16].cr_account,
                         gls.incomesummary_account.name)

        self.assertEqual(self.object.gl.transactions[17].amount, 500)
        self.assertEqual(self.object.gl.transactions[17].is_closing_cr_account,
                         True)
        self.assertEqual(self.object.gl.transactions[17].dt_account,
                         gls.incomesummary_account.name)
        self.assertEqual(self.object.gl.transactions[17].cr_account,
                         self.expa_acc.name)

        self.assertEqual(self.object.gl.transactions[18].amount, 60)
        self.assertEqual(self.object.gl.transactions[18].is_closing_cr_account,
                         True)
        self.assertEqual(self.object.gl.transactions[18].dt_account,
                         gls.gross_profit_account.name)
        self.assertEqual(self.object.gl.transactions[18].cr_account,
                         gls.incomesummary_account.name)

        self.assertEqual(self.object.gl.transactions[19].amount, 360)
        self.assertEqual(self.object.gl.transactions[19].is_closing_cr_account,
                         True)
        self.assertEqual(self.object.gl.transactions[19].dt_account,
                         gls.incomesummary_account.name)
        self.assertEqual(self.object.gl.transactions[19].cr_account,
                         gls.retainedearnings_account.name)


# =============================================================================
# Display documentation and run tests.
# =============================================================================
# os.system("cls")

# help(Component)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
This module provides a entity class representing a business entity.
an entity consists of business components e.g. Sales department.\n

@name: entity
@author: Ex Mente Technologies (Pty) Ltd
"""

from datetime import datetime
from datetime import timedelta
from dateutil.relativedelta import relativedelta
from collections import OrderedDict
from auxi.core.namedobject import NamedObject
from auxi.modeling.business.component import Component
from auxi.modeling.financial.des.generalledgeraccount import AccountType
from auxi.modeling.financial.des.generalledger import GeneralLedger
from auxi.modeling.financial.tax.ruleset import RuleSet

__version__ = "0.2.0"

# TODO: initialise accounts by json file
# TODO: get_account by path.


class Entity(NamedObject):
    """Represents an entity class."""

    # -------------------------------------------------------------------------
    # Standard methods.
    # -------------------------------------------------------------------------
    def __init__(self, name, gl_structure, description=None):
        """Initialise the object.

        :param name: The name.
        :param gl_structure: The general ledger structure the entity's general ledger will be initialized with.
        :param description: The description.
        :param period_count: The number of periods the entity should be run for.
        """
        self.gl = GeneralLedger(
            "General Ledger",
            gl_structure,
            description="General Ledger")
        self._parent_path = ""
        self.path = name
        self.components = []
        self.tax_rule_set = RuleSet("Default")
        self.negative_income_tax_total = 0
        self._prev_year_end_datetime = datetime.min
        self._curr_year_end_datetime = datetime.min
        self._exec_year_end_datetime = datetime.min
        self.period_count = -1
        super().__init__(name, description=description)

    def set_parent_path(self, value):
        """Set the parent path and the path from the new parent path.

        :param value: The path to the object's parent
        """
        self._parent_path = value
        self.path = value + r'/' + self.name
        self._update_childrens_parent_path()

    def _update_childrens_parent_path(self):
        for c in self.components:
            c.set_parent_path(self.path)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        self.path = self._parent_path + r'/' + self.name
        self._update_childrens_parent_path()

    def create_component(self, name, description=None):
        """Create a component in the business entity.

        :param name: The component's name.
        :param description: The component's description.

        :returns: The created component.
        """
        new_comp = Component(name, description=description)
        self.components.append(new_comp)
        return new_comp

    def remove_component(self, name):
        """Remove a component from the entity.

        :param name: The name of the component to remove.
        """
        component_to_remove = None
        for c in self.components:
            if c.name == name:
                component_to_remove = c
        if component_to_remove is not None:
            self.components.remove(component_to_remove)

    def prepare_to_run(self, clock, period_count):
        """Prepare the entity for execution.

        :param clock: The clock containing the execution start time and execution period information.
        :param period_count: The total amount of periods this activity will be requested to be run for.
        """
        self.period_count = period_count

        self._exec_year_end_datetime = clock.get_datetime_at_period_ix(
            period_count)
        self._prev_year_end_datetime = clock.start_datetime
        self._curr_year_end_datetime = clock.start_datetime + relativedelta(
            years=1)

        # Remove all the transactions
        del self.gl.transactions[:]

        for c in self.components:
            c.prepare_to_run(clock, period_count)

        self.negative_income_tax_total = 0

    def run(self, clock):
        """Execute the entity at the current clock cycle.

        :param clock: The clock containing the current execution time and period information.
        """
        if clock.timestep_ix >= self.period_count:
            return

        for c in self.components:
            c.run(clock, self.gl)

        self._perform_year_end_procedure(clock)

    def _perform_year_end_procedure(self, clock):
        if clock.get_datetime() >= self._curr_year_end_datetime\
         or clock.timestep_ix == self.period_count:
            gls = self.gl.structure
            year_start_date = self._prev_year_end_datetime
            year_end_date = self._curr_year_end_datetime + timedelta(seconds=-1)

            self._prev_year_end_datetime = self._curr_year_end_datetime
            self._curr_year_end_datetime += relativedelta(years=1)
            if self._curr_year_end_datetime > self._exec_year_end_datetime:
                self._curr_year_end_datetime = self._exec_year_end_datetime

            gross_profit_write_off_accs = OrderedDict()
            inc_summary_write_off_accs = OrderedDict()
            sales_accs = []
            cost_of_sales_accs = []
            gls.get_account_and_decendants(gls.sales_account, sales_accs)
            gls.get_account_and_decendants(
                gls.costofsales_account,
                cost_of_sales_accs)

            year_txes = []
            for tx in self.gl.transactions:
                if tx.tx_datetime >= year_start_date and\
                 tx.tx_datetime <= year_end_date:
                    year_txes.append(tx)
            for tx in year_txes:
                cr_acc = gls.get_account(tx.cr_account)
                dt_acc = gls.get_account(tx.dt_account)

                if cr_acc in sales_accs:
                    if cr_acc in gross_profit_write_off_accs:
                        gross_profit_write_off_accs[cr_acc] += tx.amount
                    else:
                        gross_profit_write_off_accs[cr_acc] = tx.amount
                elif dt_acc in sales_accs:
                    if dt_acc in gross_profit_write_off_accs:
                        gross_profit_write_off_accs[dt_acc] -= tx.amount
                    else:
                        gross_profit_write_off_accs[dt_acc] = -tx.amount

                elif cr_acc in cost_of_sales_accs:
                    if cr_acc in gross_profit_write_off_accs:
                        gross_profit_write_off_accs[cr_acc] += tx.amount
                    else:
                        gross_profit_write_off_accs[cr_acc] = tx.amount
                elif dt_acc in cost_of_sales_accs:
                    if dt_acc in gross_profit_write_off_accs:
                        gross_profit_write_off_accs[dt_acc] -= tx.amount
                    else:
                        gross_profit_write_off_accs[dt_acc] = -tx.amount

                elif cr_acc.account_type == AccountType.revenue:
                    if cr_acc in inc_summary_write_off_accs:
                        inc_summary_write_off_accs[cr_acc] += tx.amount
                    else:
                        inc_summary_write_off_accs[cr_acc] = tx.amount
                elif dt_acc.account_type == AccountType.revenue:
                    if dt_acc.account_type == AccountType.revenue:
                        inc_summary_write_off_accs[dt_acc] -= tx.amount
                    else:
                        inc_summary_write_off_accs[dt_acc] = -tx.amount

                elif cr_acc.account_type == AccountType.expense:
                    if cr_acc in inc_summary_write_off_accs:
                        inc_summary_write_off_accs[cr_acc] -= tx.amount
                    else:
                        inc_summary_write_off_accs[cr_acc] = -tx.amount
                elif dt_acc.account_type == AccountType.expense:
                    if dt_acc in inc_summary_write_off_accs:
                        inc_summary_write_off_accs[dt_acc] += tx.amount
                    else:
                        inc_summary_write_off_accs[dt_acc] = tx.amount

            inc_sum = self._perform_year_end_gross_profit_and_income_summary(
                year_end_date,
                gross_profit_write_off_accs,
                inc_summary_write_off_accs)
            inc_sum = self._perform_year_end_income_tax(
                year_end_date,
                inc_sum)
            self._perform_year_end_retained_earnings(
                year_end_date,
                inc_sum)

    def _perform_year_end_gross_profit_and_income_summary(
            self,
            year_end_datetime,
            gross_profit_write_off_accounts,
            income_summary_write_off_accounts):
        gls = self.gl.structure
        gross_profit = self._perform_year_end_gross_profit(
            year_end_datetime,
            gross_profit_write_off_accounts)
        income_summary_amount = self._perform_year_end_income_summary(
            year_end_datetime,
            gross_profit,
            income_summary_write_off_accounts)

        def create_tx(dt_account, cr_account):
            tx_name = "Settle the '" + gls.gross_profit_account.name +\
             "' Account"
            new_tx = self.gl.create_transaction(
                tx_name,
                description=tx_name,
                tx_datetime=year_end_datetime,
                dt_account=dt_account,
                cr_account=cr_account,
                source=self.path,
                amount=abs(gross_profit))
            new_tx.is_closing_cr_account = True

        if gross_profit > 0:
            create_tx(gls.gross_profit_account.name,
                      gls.incomesummary_account.name)
        elif gross_profit < 0:
            create_tx(gls.incomesummary_account.name,
                      gls.gross_profit_account.name)

        return income_summary_amount

    def _perform_year_end_gross_profit(self,
                                       year_end_datetime,
                                       gross_profit_write_off_accounts):
        gls = self.gl.structure
        gross_profit = 0

        for acc, amount in gross_profit_write_off_accounts.items():
            def create_tx(dt_account, cr_account):
                tx_name = "Settle the '" + acc.name + "' Account"
                new_tx = self.gl.create_transaction(
                    tx_name,
                    description=tx_name,
                    tx_datetime=year_end_datetime,
                    dt_account=dt_account,
                    cr_account=cr_account,
                    source=self.path,
                    amount=abs(amount))
                new_tx.is_closing_cr_account = True

            gross_profit += amount
            if amount > 0:
                create_tx(gls.gross_profit_account.name, acc.name)
            elif amount < 0:
                create_tx(acc.name, gls.gross_profit_account.name)
        return gross_profit

    def _perform_year_end_income_summary(self,
                                         year_end_datetime,
                                         gross_profit,
                                         income_summary_write_off_accounts):
        gls = self.gl.structure
        income_summary_amount = gross_profit

        for acc, amount in income_summary_write_off_accounts.items():
            def create_tx(dt_account, cr_account):
                tx_name = "Settle the '" + acc.name + "' Account"
                new_tx = self.gl.create_transaction(
                    tx_name,
                    description=tx_name,
                    tx_datetime=year_end_datetime,
                    dt_account=dt_account,
                    cr_account=cr_account,
                    source=self.path,
                    amount=abs(amount))
                new_tx.is_closing_cr_account = True

            if amount > 0:
                if acc.account_type == AccountType.expense:
                    income_summary_amount -= amount
                    create_tx(gls.incomesummary_account.name, acc.name)
                else:
                    income_summary_amount += amount
                    create_tx(acc.name, gls.incomesummary_account.name)
            elif amount < 0:
                if acc.account_type == AccountType.expense:
                    income_summary_amount -= amount
                    create_tx(acc.name, gls.incomesummary_account.name)
                else:
                    income_summary_amount += amount
                    create_tx(gls.incomesummary_account.name, acc.name)
        return income_summary_amount

    def _perform_year_end_income_tax(self,
                                     year_end_datetime,
                                     income_summary_amount):
        return income_summary_amount

    def _perform_year_end_retained_earnings(self,
                                            year_end_datetime,
                                            income_summary_amount):
        def create_tx(dt_account, cr_account):
            tx_name = "Settle the '" + gls.incomesummary_account.name +\
                      "' Account, adjust Retained Earnings accordingly."
            new_tx = self.gl.create_transaction(
                tx_name,
                description=tx_name,
                tx_datetime=year_end_datetime,
                dt_account=dt_account,
                cr_account=cr_account,
                source=self.path,
                amount=income_summary_amount)
            new_tx.is_closing_cr_account = True
        gls = self.gl.structure

        if income_summary_amount > 0:
            create_tx(gls.incomesummary_account.name,
                      gls.retainedearnings_account.name)
        elif income_summary_amount < 0:
            create_tx(gls.retainedearnings_account.name,
                      gls.incomesummary_account.name)



# -*- coding: utf-8 -*-
"""
This module provides the base class for assay conversion actions.

@author: Ex Mente Technologies (Pty) Ltd
"""
__version__ = "0.2.0"


class AssayConversionAction(object):
    def run(self, material):
        pass



# -*- coding: utf-8 -*-
"""
This module provides a number of functions for doing stoichiometry
calculations.

@author: Ex Mente Technologies (Pty) Ltd
"""

import numpy
from auxi.tools.chemistry.element import Element

__version__ = "0.2.0"


# =============================================================================
# Private functions.
# =============================================================================

def _formula_code(formula=str()):
    """Generate a unique formula code for a chemical compound.

    :param formula: The formula describing the chemical compound.

    :returns: The formula code for the chemical compouned.
    """

    result = str()
    if len(formula) == 0:
        return result

    codeSum = 0
    for i in range(len(formula)):
        c = formula[i:i+1]
        b = ord(c)
        result = result + str(b)
        codeSum = codeSum + b
    result = result + "_" + str(codeSum)

    return result


def _get_character(string, index=0):
    """Returns the character from a string at the index position, the integer
    ordinal of the character, and an incremented index.

    If index is at the end of the string, return an empty string, -1 for the \
    ordinal, and return index unchanged.

    :param string: String to search.
    :param index:  Index at which the character should be located.

    :returns: character at string[index]
    :returns: ordinal of character
    :returns: incremented index"""

    if index == len(string):
        return "", -1, index
    else:
        character = string[index:index+1]
        ordinal = ord(character)
        return character, ordinal, index + 1


def _get_formula(compound):
    """Remove the phase from a compound string if it exists and return only
    the formula.

    :param compound: Formula and phase of a chemical compound, e.g. \
    "Fe2O3[S1]".

    :returns: Chemical formula.
    """

    return compound.split("[")[0]


def _parse_element_for_mass(compound, index):
    """Determine the atomic mass of the element at the specified index of the
    chemical compound formula by parsing the formula.

    :param compound: Formula of a chemical compound, "Fe3O4".
    :param index:    Index from which the formula should be parsed.

    :returns: Atomic mass. [kg/kmol]
    """

    Element = compound[index:index+1]
    index = index + 1
    if index < len(compound):
        code = ord(compound[index:index+1])
    else:
        code = 0

    while code >= 97 and code <= 123:
        Element = Element + compound[index:index+1]
        index = index + 1
        if index == len(compound):
            code = 0
            break
        code = ord(compound[index:index+1])

    multiplier = str()
    while (code >= 48 and code <= 57) or code == 46:
        multiplier = multiplier + compound[index:index+1]
        index = index + 1
        if index == len(compound):
            break
        code = ord(compound[index:index+1])
    if multiplier == "":
        multiplier = "1"

    result = _element_dictionary[Element].molar_mass * float(multiplier)
    return (result, index)


def _parse_element_for_stoichiometry(compound, index):
    """Determine the stoichiometry coefficient of the element at the specified
    index of the formula by parsing the formula.

    :param compound: Formula of a chemical compound, "Fe3O4".
    :param index:    Index from which the formula should be parsed.

    :returns: Stoichiometric coefficient.
    """

    element_symbol = compound[index:index+1]
    index = index + 1
    if index < len(compound):
        c = compound[index:index+1]
        b = ord(c)
    else:
        b = 0

    while (b >= 97 and b <= 123):
        element_symbol = element_symbol + compound[index:index+1]
        index = index + 1
        if index >= len(compound):
            return (element_symbol, 1.0, index)
        c = compound[index:index+1]
        b = ord(c)

    stoichiometry_coefficient = ""

    while ((b >= 48 and b <= 57) or b == 46):
        stoichiometry_coefficient = stoichiometry_coefficient + compound[index:index+1]
        index = index + 1
        if index >= len(compound):
            break
        c = compound[index:index+1]
        b = ord(c)

    if stoichiometry_coefficient == "":
        stoichiometry_coefficient = "1.0"

    return (element_symbol, float(stoichiometry_coefficient), index)


def _parse_formula_for_elements(compound):
    """Determine the set of elements that occur in the specified formula.

    :param compound: Formula of a chemical compound.

    :returns: Set of elements.
    """

    # Initialise the search variables.
    result = set()
    i = 0              # The index of the current character in the string.

    # Do the search.
    while i < len(compound):
        c, b, i = _get_character(compound, i)
        if b >= 65 and b <= 90:  # Element found. Process it.
            j = i
            element = c
            c, b, j = _get_character(compound, j)
            while b >= 97 and b <= 122:
                element = element + c
                c, b, j = _get_character(compound, j)
            result.add(element)

    return result


def _parse_formula_for_mass(compound, index):
    """Determine the molar mass of the chemical compound by recursively
    parsing its formula from the specified index.

    :param compound: Formula of chemical compound.
    :param index:    Index from which the formula should be parsed.

    :returns: Molecular mass of the parsed portion of the formula.
    """
    result = 0.0

    c = str()
    while index < len(compound) and c != ")":
        c = compound[index:index+1]
        b = ord(c)

        if c == "(":
            index = index + 1
            dresult, index = _parse_formula_for_mass(compound, index)
            result = result + dresult
        elif b >= 65 and b <= 90:
            dresult, index = _parse_element_for_mass(compound, index)
            result = result + dresult

    if index >= len(compound):
        return result

    if c == ")":
        index = index + 1
    b = ord(compound[index:index+1])

    multiplier = str()
    while ((b >= 48 and b <= 57) or b == 46):
        multiplier = multiplier + compound[index:index+1]
        index = index + 1
        if index == len(compound):
            break
        c = compound[index:index+1]
        b = ord(c)

    if multiplier == "":
        multiplier = "1"

    result = result * float(multiplier)
    return (result, index)


def _parse_formula_for_stoichiometry(compound, index, stoich_dict):
    """Determine the stoichiometry of the formula by recursively parsing it
    from the specified index and storing the result in the specified dictionary.

    :param compound:    Formula of chemical compound.
    :param index:       Index from which the formula should be parsed.
    :param stoich_dict: Stoichiometry dictionary.
    """
    c = str()
    while index < len(compound) and c != ")":
        c = compound[index:index+1]
        b = ord(c)

        if c == "(":
            index = index + 1
            new_stoichiometry_records = {}
            _parse_formula_for_stoichiometry(compound, index, new_stoichiometry_records)
            for k, v in new_stoichiometry_records.items():
                if k in stoich_dict:
                    stoich_dict[k] = stoich_dict[k] + new_stoichiometry_records[k]
                else:
                    stoich_dict[k] = v
        else:
            if b >= 65 and b <= 90:
                element, coefficient, index = _parse_element_for_stoichiometry(compound, index)
                if element in stoich_dict:
                    stoich_dict[element] = stoich_dict[element] + coefficient
                else:
                    stoich_dict[element] = coefficient

    if index >= len(compound):
        return

    if c == ")":
        index = index + 1
    c = compound[index:index+1]
    b = ord(c)

    multplier_string = str()
    while ((b >= 48 and b <= 57) or b == 46):
        multplier_string = multplier_string + compound[index:index+1]
        index = index + 1
        if index == len(compound):
            return
        c = compound[index:index+1]
        b = ord(c)

    if multplier_string != "":
        multiplier = float(multplier_string)
        for k, v in stoich_dict.items():
            stoich_dict[k] = stoich_dict[k] * multiplier


def _populate_element_dictionary():
    """Create all the elements of the periodic table and add them to the
    element dictionary."""

    # period 1
    _element_dictionary["H"]   = Element(1,  1,   1, "H",    1.00794)
    _element_dictionary["He"]  = Element(1, 18,   2, "He",   4.002602)

    # period 2
    _element_dictionary["Li"]  = Element(2,  1,   3, "Li",   6.941)
    _element_dictionary["Be"]  = Element(2,  2,   4, "Be",   9.012182)
    _element_dictionary["B"]   = Element(2, 13,   5, "B",   10.811)
    _element_dictionary["C"]   = Element(2, 14,   6, "C",   12.0107)
    _element_dictionary["N"]   = Element(2, 15,   7, "N",   14.00674)
    _element_dictionary["O"]   = Element(2, 16,   8, "O",   15.9994)
    _element_dictionary["F"]   = Element(2, 17,   9, "F",   18.9984032)
    _element_dictionary["Ne"]  = Element(2, 18,  10, "Ne",  20.1797)

    # period 3
    _element_dictionary["Na"]  = Element(3,  1,  11, "Na",  22.98977)
    _element_dictionary["Mg"]  = Element(3,  2,  12, "Mg",  24.305)
    _element_dictionary["Al"]  = Element(3, 13,  13, "Al",  26.981538)
    _element_dictionary["Si"]  = Element(3, 14,  14, "Si",  28.0855)
    _element_dictionary["P"]   = Element(3, 15,  15, "P",   30.973762)
    _element_dictionary["S"]   = Element(3, 16,  16, "S",   32.066)
    _element_dictionary["Cl"]  = Element(3, 17,  17, "Cl",  35.4527)
    _element_dictionary["Ar"]  = Element(3, 18,  18, "Ar",  39.948)

    # period 4
    _element_dictionary["K"]   = Element(4,  1,  19, "K",   39.0983)
    _element_dictionary["Ca"]  = Element(4,  2,  20, "Ca",  40.078)
    _element_dictionary["Sc"]  = Element(4,  3,  21, "Sc",  44.95591)
    _element_dictionary["Ti"]  = Element(4,  4,  22, "Ti",  47.867)
    _element_dictionary["V"]   = Element(4,  5,  23, "V",   50.9415)
    _element_dictionary["Cr"]  = Element(4,  6,  24, "Cr",  51.9961)
    _element_dictionary["Mn"]  = Element(4,  7,  25, "Mn",  54.938049)
    _element_dictionary["Fe"]  = Element(4,  8,  26, "Fe",  55.845)
    _element_dictionary["Co"]  = Element(4,  9,  27, "Co",  58.9332)
    _element_dictionary["Ni"]  = Element(4, 10,  28, "Ni",  58.6934)
    _element_dictionary["Cu"]  = Element(4, 11,  29, "Cu",  63.546)
    _element_dictionary["Zn"]  = Element(4, 12,  30, "Zn",  65.39)
    _element_dictionary["Ga"]  = Element(4, 13,  31, "Ga",  69.723)
    _element_dictionary["Ge"]  = Element(4, 14,  32, "Ge",  72.61)
    _element_dictionary["As"]  = Element(4, 15,  33, "As",  74.9216)
    _element_dictionary["Se"]  = Element(4, 16,  34, "Se",  78.96)
    _element_dictionary["Br"]  = Element(4, 17,  35, "Br",  79.904)
    _element_dictionary["Kr"]  = Element(4, 18,  36, "Kr",  83.8)

    # period 5
    _element_dictionary["Rb"]  = Element(5,  1,  37, "Rb",  85.4678)
    _element_dictionary["Sr"]  = Element(5,  2,  38, "Sr",  87.62)
    _element_dictionary["Y"]   = Element(5,  3,  39, "Y",   88.90585)
    _element_dictionary["Zr"]  = Element(5,  4,  40, "Zr",  91.224)
    _element_dictionary["Nb"]  = Element(5,  5,  41, "Nb",  92.90638)
    _element_dictionary["Mo"]  = Element(5,  6,  42, "Mo",  95.94)
    _element_dictionary["Tc"]  = Element(5,  7,  43, "Tc",  98.0)
    _element_dictionary["Ru"]  = Element(5,  8,  44, "Ru",  101.07)
    _element_dictionary["Rh"]  = Element(5,  9,  45, "Rh",  102.9055)
    _element_dictionary["Pd"]  = Element(5, 10,  46, "Pd",  106.42)
    _element_dictionary["Ag"]  = Element(5, 11,  47, "Ag",  107.8682)
    _element_dictionary["Cd"]  = Element(5, 12,  48, "Cd",  112.411)
    _element_dictionary["In"]  = Element(5, 13,  49, "In",  114.818)
    _element_dictionary["Sn"]  = Element(5, 14,  50, "Sn",  118.71)
    _element_dictionary["Sb"]  = Element(5, 15,  51, "Sb",  121.76)
    _element_dictionary["Te"]  = Element(5, 16,  52, "Te",  127.6)
    _element_dictionary["I"]   = Element(5, 17,  53, "I",   126.90447)
    _element_dictionary["Xe"]  = Element(5, 18,  54, "Xe",  131.29)

    # period 6
    _element_dictionary["Cs"]  = Element(6,  1,  55, "Cs",  132.90545)
    _element_dictionary["Ba"]  = Element(6,  2,  56, "Ba",  137.327)
    _element_dictionary["La"]  = Element(6,  0,  57, "La",  138.9055)
    _element_dictionary["Ce"]  = Element(6,  0,  58, "Ce",  140.116)
    _element_dictionary["Pr"]  = Element(6,  0,  59, "Pr",  140.90765)
    _element_dictionary["Nd"]  = Element(6,  0,  60, "Nd",  144.24)
    _element_dictionary["Pm"]  = Element(6,  0,  61, "Pm",  145.0)
    _element_dictionary["Sm"]  = Element(6,  0,  62, "Sm",  150.36)
    _element_dictionary["Eu"]  = Element(6,  0,  63, "Eu",  151.964)
    _element_dictionary["Gd"]  = Element(6,  0,  64, "Gd",  157.25)
    _element_dictionary["Tb"]  = Element(6,  0,  65, "Tb",  158.92534)
    _element_dictionary["Dy"]  = Element(6,  0,  66, "Dy",  162.5)
    _element_dictionary["Ho"]  = Element(6,  0,  67, "Ho",  164.93032)
    _element_dictionary["Er"]  = Element(6,  0,  68, "Er",  167.26)
    _element_dictionary["Tm"]  = Element(6,  0,  69, "Tm",  168.93421)
    _element_dictionary["Yb"]  = Element(6,  0,  70, "Yb",  173.04)
    _element_dictionary["Lu"]  = Element(6,  0,  71, "Lu",  174.967)
    _element_dictionary["Hf"]  = Element(6,  4,  72, "Hf",  178.49)
    _element_dictionary["Ta"]  = Element(6,  5,  73, "Ta",  180.9479)
    _element_dictionary["W"]   = Element(6,  6,  74, "W",   183.84)
    _element_dictionary["Re"]  = Element(6,  7,  75, "Re",  186.207)
    _element_dictionary["Os"]  = Element(6,  8,  76, "Os",  190.23)
    _element_dictionary["Ir"]  = Element(6,  9,  77, "Ir",  192.217)
    _element_dictionary["Pt"]  = Element(6, 10,  78, "Pt",  195.078)
    _element_dictionary["Au"]  = Element(6, 11,  79, "Au",  196.96655)
    _element_dictionary["Hg"]  = Element(6, 12,  80, "Hg",  200.59)
    _element_dictionary["Tl"]  = Element(6, 13,  81, "Tl",  204.3833)
    _element_dictionary["Pb"]  = Element(6, 14,  82, "Pb",  207.2)
    _element_dictionary["Bi"]  = Element(6, 15,  83, "Bi",  208.98038)
    _element_dictionary["Po"]  = Element(6, 16,  84, "Po",  210.0)
    _element_dictionary["At"]  = Element(6, 17,  85, "At",  210.0)
    _element_dictionary["Rn"]  = Element(6, 18,  86, "Rn",  222.0)

    # period 7
    _element_dictionary["Fr"]  = Element(7,  1, 87, "Fr",   223.0)
    _element_dictionary["Ra"]  = Element(7,  2, 88, "Ra",   226.0)
    _element_dictionary["Ac"]  = Element(7,  0, 89, "Ac",   227.0)
    _element_dictionary["Th"]  = Element(7,  0, 90, "Th",   232.0381)
    _element_dictionary["Pa"]  = Element(7,  0, 91, "Pa",   231.03588)
    _element_dictionary["U"]   = Element(7,  0, 92, "U",    238.0289)
    _element_dictionary["Np"]  = Element(7,  0, 93, "Np",   237.0)
    _element_dictionary["Pu"]  = Element(7,  0, 94, "Pu",   244.0)
    _element_dictionary["Am"]  = Element(7,  0, 95, "Am",   243.0)
    _element_dictionary["Cm"]  = Element(7,  0, 96, "Cm",   247.0)
    _element_dictionary["Bk"]  = Element(7,  0, 97, "Bk",   247.0)
    _element_dictionary["Cf"]  = Element(7,  0, 98, "Cf",   251.0)
    _element_dictionary["Es"]  = Element(7,  0, 99, "Es",   252.0)
    _element_dictionary["Fm"]  = Element(7,  0, 100, "Fm",  257.0)
    _element_dictionary["Md"]  = Element(7,  0, 101, "Md",  258.0)
    _element_dictionary["No"]  = Element(7,  0, 102, "No",  259.0)
    _element_dictionary["Lr"]  = Element(7,  0, 103, "Lr",  262.0)
    _element_dictionary["Rf"]  = Element(7,  4, 104, "Rf",  261.0)
    _element_dictionary["Db"]  = Element(7,  5, 105, "Db",  262.0)
    _element_dictionary["Sg"]  = Element(7,  6, 106, "Sg",  266.0)
    _element_dictionary["Bh"]  = Element(7,  7, 107, "Bh",  264.0)
    _element_dictionary["Hs"]  = Element(7,  8, 108, "Hs",  269.0)
    _element_dictionary["Mt"]  = Element(7,  9, 109, "Mt",  268.0)
    _element_dictionary["Ds"]  = Element(7, 10, 110, "Ds",  269.0)
    _element_dictionary["Rg"]  = Element(7, 11, 111, "Rg",  272.0)
    # Cn missing
    # Uut missing
    # Fl missing
    # Uup missing
    # Lv missing
    # Uus missing
    # Uuo missing
    # actinides


# =============================================================================
# Public functions.
# =============================================================================

def amount(compound, mass):
    """Calculate the number of moles in the specified mass of a chemical
    compound.

    :param compound: Formula and phase of a compound, e.g. "Fe2O3[S1]". The \
    phase may be omitted.
    :param mass:     [kg]

    :returns: Amount. [kmol]
    """

    return mass / molar_mass(_get_formula(compound))


def mass(compound, amount):
    """Calculate the mass of the specified amount of a chemical compound.

    :param compound: Formula and phase of a compound, e.g. "Fe2O3[S1]". The \
phase may be omitted.
    :param amount:   [kmol]

    :returns: Mass. [kg]
    """

    return amount * molar_mass(_get_formula(compound))


def convert_compound(mass, source, target, element):
    """Convert the specified mass of the source compound to the target using
    element as basis.

    :param mass:    Mass of from_compound. [kg]
    :param source:  Formula and phase of the original compound, e.g. \
    "Fe2O3[S1]".
    :param target:  Formula and phase of the target compound, e.g. \
    "Fe[S1]".
    :param element: Element to use as basis for the conversion, e.g. \
    "Fe" or "O".

    :returns: Mass of to_compound. [kg]
    """

    # Convert compounds to formulas.
    source_formula = _get_formula(source)
    target_formula = _get_formula(target)

    # Perform the conversion.
    source_mass_fraction = element_mass_fraction(source_formula, element)
    target_mass_fraction = element_mass_fraction(target_formula, element)
    if target_mass_fraction == 0.0:
        # If target_formula does not contain element, just return 0.0.
        return 0.0
    else:
        return mass * source_mass_fraction / target_mass_fraction


def element_mass_fraction(compound, element):
    """Determine the mass fraction of an element in a chemical compound.

    :param compound: Formula of the chemical compound, "FeCr2O4".
    :param element: Element, e.g. "Cr".

    :returns: Element mass fraction."""

    elementStoichiometryCoefficient = stoichiometry_coefficient(compound, element)
    if elementStoichiometryCoefficient == 0.0:
        return 0.0
    else:
        formulaMass = molar_mass(compound)
        elementMass = molar_mass(element)
        return elementStoichiometryCoefficient * elementMass / formulaMass


def element_mass_fractions(compound, elements):
    """Determine the mass fractions of a list of elements in a chemical
    compound.

    :param compound: Formula and phase of a chemical compound, e.g. \
    "Fe2O3[S1]".
    :param elements: List of elements, ["Si", "O", "Fe"].

    :returns: Mass fractions."""

    formula = _get_formula(compound)
    result = numpy.array([])
    for i in range(0, len(elements)):
        result = numpy.append(result, element_mass_fraction(formula, elements[i]))
    return result


def elements(compounds):
    """Determine the set of elements present in a list of chemical compounds.

    The list of elements is sorted alphabetically.

    :param compounds: List of compound formulas and phases, e.g. \
    ["Fe2O3[S1]", "Al2O3[S1]"].

    :returns: List of elements.
    """

    result = set()
    for compound in compounds:
        formula = _get_formula(compound)
        result = result.union(_parse_formula_for_elements(formula))
    return result


def molar_mass(compound=""):
    """Determine the molar mass of a chemical compound.

    The molar mass is usually the mass of one mole of the substance, but here
    it is the mass of 1000 moles, since the mass unit used in pmpy is kg.

    :param compound: Formula of a chemical compound, e.g. "Fe2O3".

    :returns: Molar mass. [kg/kmol]"""

    result = 0.0
    compound = compound.strip()
    if len(compound) == 0:
        return result

    code = _formula_code(compound)
    if code not in _molar_mass_dictionary:
        index = 0
        _molar_mass_dictionary[code] = _parse_formula_for_mass(compound, index)
    result = _molar_mass_dictionary[code]

    return result


def stoichiometry_coefficient(compound, element):
    """Determine the stoichiometry coefficient of an element in a chemical
    compound.

    :param compound: Formula of a chemical compound, e.g. "SiO2".
    :param element:  Element, e.g. "Si".

    :returns: Stoichiometry coefficient."""

    compound = compound.strip()

    if compound not in _stoichiometry_dictionary:
        stoichiometry = {}
        _parse_formula_for_stoichiometry(compound, 0, stoichiometry)
        _stoichiometry_dictionary[_formula_code(compound)] = stoichiometry

    stoichiometry = _stoichiometry_dictionary[_formula_code(compound)]

    if element in stoichiometry:
        return stoichiometry[element]
    else:
        return 0.0


# TODO: Fix problem when there are brackets in the formula.
def stoichiometry_coefficients(compound, elements):
    """Determine the stoichiometry coefficients of the specified elements in
    the specified chemical compound.

    :param compound: Formula of a chemical compound, e.g. "SiO2".
    :param elements: List of elements, e.g. ["Si", "O", "C"].

    :returns: List of stoichiometry coefficients."""

    result = []
    for i in range(0, len(elements)):
        result.append(stoichiometry_coefficient(compound, elements[i]))
    return result


# =============================================================================
# Private variables.
# =============================================================================

_element_dictionary = {}
_molar_mass_dictionary = {}
_stoichiometry_dictionary = {}


# =============================================================================
# Module initialisation.
# =============================================================================

_populate_element_dictionary()



# -*- coding: utf-8 -*-
"""
This module provides testing code for the modelling.thermochemistry module.

@author: Ex Mente Technologies (Pty) Ltd
"""

# TODO: material package uses materials 'assays' property.
#       Material does not have an 'assays' property,
#       only a raw_assays and converted_assays property.

import unittest
import os
import numpy
from auxi.modeling.process.materials.thermochemistry import material
from auxi.modeling.process.materials.thermochemistry.material import Material
from auxi.modeling.process.materials.thermochemistry.material import MaterialPackage
from auxi.tools.chemistry import thermochemistry as thermo

__version__ = "0.2.0"

# =============================================================================
# Types.
# =============================================================================


class TestMaterial(unittest.TestCase):
    """Tester for the auxi.modeling.process.materials.thermochemistry.material.Material class."""
    def setUp(self):
        self.material = Material("material", os.path.join(material.DEFAULT_DATA_PATH, r"thermomaterial.test.ilmenite.txt"))

    def test_constructor(self):
        self.assertEqual(self.material.name, "material")
        self.assertEqual(len(self.material.compounds), 14)
        self.assertEqual(self.material.compound_count, 14)
        self.assertEqual(len(self.material.assays), 3)

    def test_get_compound_index(self):
        self.assertEqual(self.material.get_compound_index("Al2O3[S1]"), 0)
        self.assertEqual(self.material.get_compound_index("K2O[S]"), 6)
        self.assertEqual(self.material.get_compound_index("P4O10[S]"), 10)
        self.assertEqual(self.material.get_compound_index("V2O5[S]"), 13)

    def test_create_empty_assay(self):
        empty_assay = self.material.create_empty_assay()
        self.assertEqual(len(empty_assay), 14)
        self.assertEqual(empty_assay.sum(), 0.0)
        self.assertEqual(empty_assay.sum(), 0.0)
        self.assertEqual(empty_assay.sum(), 0.0)

    def test_add_assay(self):
        new_assay = self.material.create_empty_assay()
        new_assay[0] = 0.5
        new_assay[2] = 0.5
        self.material.add_assay("new_assay", new_assay)
        self.assertEqual(numpy.all(self.material.assays["new_assay"] == new_assay), True)

    def test_get_assay_total(self):
        self.assertEqual(self.material.get_assay_total("IlmeniteA"), 0.99791999999999992)
        self.assertEqual(self.material.get_assay_total("IlmeniteB"), 0.99760999999999989)
        self.assertEqual(self.material.get_assay_total("IlmeniteC"), 1.0000200000000001)

    def test_create_package(self):
        package = self.material.create_package("IlmeniteA", 123.456, 0.87, 205.0, True)
        self.assertEqual(package.get_mass(), 123.456)
        self.assertEqual(package.P, 0.87)
        self.assertEqual(package.T, 205.0)
        self.assertEqual(package.H, -278.35680682442677)


class TestMaterialPackage(unittest.TestCase):
    """Tester for the auxi.modeling.process.materials.thermochemistry.material.MaterialPackage class."""
    def setUp(self):
        self.ilmenite = Material("ilmenite", os.path.join(material.DEFAULT_DATA_PATH, r"thermomaterial.test.ilmenite.txt"))
        self.ilmenitePackageA = self.ilmenite.create_package("IlmeniteA", 1234.5, 0.8, 100.0, True)
        self.ilmenitePackageB = self.ilmenite.create_package("IlmeniteB", 2345.6, 0.9, 200.0, True)
        self.ilmenitePackageC = self.ilmenite.create_package("IlmeniteC", 3456.7, 1.0, 300.0, True)
        self.reductant = Material("reductant", os.path.join(material.DEFAULT_DATA_PATH, r"thermomaterial.test.reductant.txt"))
        self.reductantPackageA = self.reductant.create_package("ReductantA", 123.45, 0.75, 400.0, True)
        self.mix = Material("mix", os.path.join(material.DEFAULT_DATA_PATH, r"thermomaterial.test.mix.txt"))

    def test_constructor(self):
        compound_masses = self.ilmenite.assays["IlmeniteB"] * 123.4 / self.ilmenite.assays["IlmeniteB"].sum()
        package = MaterialPackage(self.ilmenite, compound_masses, 0.8, 300.0)
        self.assertEqual(package.get_mass(), 123.4)
        self.assertEqual(package.P, 0.8)
        self.assertEqual(package.T, 300.0)
        self.assertEqual(package.H, -236.60618829889296)

    def test_add_operator_1(self):
        """other = MaterialPackage
        Test whether the add operator calculates the resulting package correctly.
        Results were checked against FactSage results. They are not exactly the same, since the magnetic and other non-cp contributions are omitted by the thermo module.
        """

        packageAplusB = self.ilmenitePackageA + self.ilmenitePackageB
        packageAplusC = self.ilmenitePackageA + self.ilmenitePackageC
        packageBplusC = self.ilmenitePackageB + self.ilmenitePackageC
        packageAplusBplusC = self.ilmenitePackageA + self.ilmenitePackageB + self.ilmenitePackageC

        self.assertAlmostEqual(self.ilmenitePackageA.get_mass(), 1234.5)
        self.assertAlmostEqual(self.ilmenitePackageA.P, 0.8)
        self.assertAlmostEqual(self.ilmenitePackageA.T, 100.0)
        self.assertAlmostEqual(self.ilmenitePackageA.H, -2811.6976363963095)

        self.assertAlmostEqual(self.ilmenitePackageB.get_mass(), 2345.6)
        self.assertAlmostEqual(self.ilmenitePackageB.P, 0.9)
        self.assertAlmostEqual(self.ilmenitePackageB.T, 200.0)
        self.assertAlmostEqual(self.ilmenitePackageB.H, -4550.3197620429773)

        self.assertAlmostEqual(self.ilmenitePackageC.get_mass(), 3456.7)
        self.assertAlmostEqual(self.ilmenitePackageC.P, 1.0)
        self.assertAlmostEqual(self.ilmenitePackageC.T, 300.0)
        self.assertAlmostEqual(self.ilmenitePackageC.H, -8023.545629913353)

        self.assertAlmostEqual(packageAplusB.get_mass(), 3580.0999999999999)
        self.assertAlmostEqual(packageAplusB.P, 0.8)
        self.assertAlmostEqual(packageAplusB.T, 165.93941355722774)
        self.assertAlmostEqual(packageAplusB.H, self.ilmenitePackageA.H + self.ilmenitePackageB.H)

        self.assertAlmostEqual(packageAplusC.get_mass(), 4691.1999999999989)
        self.assertAlmostEqual(packageAplusC.P, 0.8)
        self.assertAlmostEqual(packageAplusC.T, 249.48206641333098)
        self.assertAlmostEqual(packageAplusC.H, self.ilmenitePackageA.H + self.ilmenitePackageC.H)

        self.assertAlmostEqual(packageBplusC.get_mass(), 5802.2999999999993)
        self.assertAlmostEqual(packageBplusC.P, 0.9)
        self.assertAlmostEqual(packageBplusC.T, 260.56905390294497)
        self.assertAlmostEqual(packageBplusC.H, self.ilmenitePackageB.H + self.ilmenitePackageC.H)

        self.assertAlmostEqual(packageAplusBplusC.get_mass(), 7036.8)
        self.assertAlmostEqual(packageAplusBplusC.P, 0.8)
        self.assertAlmostEqual(packageAplusBplusC.T, 233.3038145723261)
        self.assertAlmostEqual(packageAplusBplusC.H, self.ilmenitePackageA.H + self.ilmenitePackageB.H + self.ilmenitePackageC.H)

    def test_add_operator_2(self):
        mixPackage = self.mix.create_package(None, 0.0)
        mixPackage = mixPackage + self.ilmenitePackageA
        mixPackage = mixPackage + self.reductantPackageA

        self.assertEqual(mixPackage.get_mass(), self.ilmenitePackageA.get_mass() + self.reductantPackageA.get_mass())
        self.assertEqual(mixPackage.P, 1.0)
        self.assertAlmostEqual(mixPackage.T, 146.64409455076031)
        self.assertEqual(mixPackage.H, self.ilmenitePackageA.H + self.reductantPackageA.H)

        self.assertRaises(Exception, self.add_incompatible_packages)

    def add_incompatible_packages(self):
        result = self.ilmenitePackageA + self.reductantPackageA
        result = result * 1.0

    def test_add_operator_3(self):
        """other = tuple (compound, mass)
        Test whether the add operator calculates the resulting package correctly.
        Results were checked against FactSage results. They are not exactly the same, since the magnetic and other non-cp contributions are omitted by the thermo module.
        """

        packageAplusAl2O3 = self.ilmenitePackageA + ("Al2O3[S1]", 123.4)

        self.assertEqual(packageAplusAl2O3.get_mass(), 1357.9)
        self.assertEqual(packageAplusAl2O3.P, 0.8)
        self.assertEqual(packageAplusAl2O3.T, 100.0)
        self.assertEqual(packageAplusAl2O3.H, self.ilmenitePackageA.H + thermo.H("Al2O3[S1]", 100.0, 123.4))

    def test_add_operator_4(self):
        """other = tuple (compound, mass, temperature)
        Test whether the add operator calculates the resulting package correctly.
        Results were checked against FactSage results. They are not exactly the same, since the magnetic and other non-cp contributions are omitted by the thermo module.
        """

        packageAplusAl2O3 = self.ilmenitePackageA + ("Al2O3[S1]", 123.4, 500.0)

        self.assertEqual(packageAplusAl2O3.get_mass(), 1357.9)
        self.assertEqual(packageAplusAl2O3.P, 0.8)
        self.assertEqual(packageAplusAl2O3.T, 151.60776535105211)
        self.assertEqual(packageAplusAl2O3.H, self.ilmenitePackageA.H + thermo.H("Al2O3[S1]", 500.0, 123.4))

    def test_extract_1(self):
        tempPackageA = self.ilmenitePackageA.clone()
        mass = 432.1
        diffPackage = tempPackageA.extract(mass)

        self.assertEqual(tempPackageA.get_mass(), self.ilmenitePackageA.get_mass() - mass)
        self.assertEqual(tempPackageA.P, 0.8)
        self.assertEqual(tempPackageA.T, 100.0)

        self.assertEqual(diffPackage.get_mass(), mass + 8.0E-14)
        self.assertEqual(diffPackage.P, 0.8)
        self.assertEqual(diffPackage.T, 100.0)

        self.assertEqual(tempPackageA.H + diffPackage.H, self.ilmenitePackageA.H)

    def test_extract_2(self):
        tempPackageA = self.ilmenitePackageA.clone()
        compound = "TiO2[S1]"
        mass = 123.4
        diffPackage = tempPackageA.extract((compound, mass))

        self.assertEqual(tempPackageA.get_mass(), self.ilmenitePackageA.get_mass() - mass + 2.0E-13)
        self.assertEqual(tempPackageA.get_compound_mass(compound), self.ilmenitePackageA.get_compound_mass(compound) - mass)
        self.assertEqual(tempPackageA.P, 0.8)
        self.assertEqual(tempPackageA.T, 100.0)

        self.assertEqual(diffPackage.get_mass(), mass)
        self.assertEqual(diffPackage.get_compound_mass(compound), mass)
        self.assertEqual(diffPackage.P, 0.8)
        self.assertEqual(diffPackage.T, 100.0)

        self.assertEqual(tempPackageA.H + diffPackage.H, self.ilmenitePackageA.H)

    def test_extract_3(self):
        tempPackageA = self.ilmenitePackageA.clone()
        compound = "TiO2[S1]"
        mass = tempPackageA.get_compound_mass(compound)
        diffPackage = tempPackageA.extract(compound)

        self.assertEqual(tempPackageA.get_mass(), self.ilmenitePackageA.get_mass() - mass + 1.0E-13)
        self.assertEqual(tempPackageA.get_compound_mass(compound), self.ilmenitePackageA.get_compound_mass(compound) - mass)
        self.assertEqual(tempPackageA.P, 0.8)
        self.assertEqual(tempPackageA.T, 100.0)

        self.assertEqual(diffPackage.get_mass(), mass)
        self.assertEqual(diffPackage.get_compound_mass(compound), mass)
        self.assertEqual(diffPackage.P, 0.8)
        self.assertEqual(diffPackage.T, 100.0)

        self.assertEqual(tempPackageA.H + diffPackage.H, self.ilmenitePackageA.H)

    def test_multiply_operator(self):
        tempPackageA = self.ilmenitePackageA.clone()

        mul1Package = tempPackageA * 0.0
        self.assertEqual(mul1Package.get_mass(), 0.0)
        self.assertTrue(numpy.all(mul1Package._compound_masses == tempPackageA._compound_masses * 0.0))
        self.assertEqual(mul1Package.P, tempPackageA.P)
        self.assertEqual(mul1Package.T, tempPackageA.T)
        self.assertEqual(mul1Package.H, 0.0)

        mul2Package = tempPackageA * 1.0
        self.assertEqual(mul2Package.get_mass(), tempPackageA.get_mass())
        self.assertTrue(numpy.all(mul2Package._compound_masses == tempPackageA._compound_masses))
        self.assertEqual(mul2Package.P, tempPackageA.P)
        self.assertEqual(mul2Package.T, tempPackageA.T)
        self.assertEqual(mul2Package.H, tempPackageA.H)

        mul2Package = tempPackageA * 123.4
        self.assertEqual(mul2Package.get_mass(), tempPackageA.get_mass() * 123.4 + 3.0E-11)
        self.assertTrue(numpy.all(mul2Package._compound_masses == tempPackageA._compound_masses * 123.4))
        self.assertEqual(mul2Package.P, tempPackageA.P)
        self.assertEqual(mul2Package.T, tempPackageA.T)
        self.assertEqual(mul2Package.H, tempPackageA.H * 123.4 - 6.0E-11)

    def test_clone(self):
        clone = self.ilmenitePackageA.clone()

        self.assertEqual(clone.get_mass(), self.ilmenitePackageA.get_mass())
        self.assertTrue(numpy.all(clone._compound_masses == self.ilmenitePackageA._compound_masses))
        self.assertEqual(clone.P, self.ilmenitePackageA.P)
        self.assertEqual(clone.T, self.ilmenitePackageA.T)
        self.assertEqual(clone.H, self.ilmenitePackageA.H)

    def test_get_mass(self):
        self.assertEqual(self.ilmenitePackageA.get_mass(), 1234.5)
        self.assertEqual(self.ilmenitePackageB.get_mass(), 2345.6)
        self.assertEqual(self.ilmenitePackageC.get_mass(), 3456.7)

    def test_get_assay(self):
        self.assertTrue(numpy.all(self.ilmenitePackageA.get_assay() - self.ilmenite.assays["IlmeniteA"] / self.ilmenite.assays["IlmeniteA"].sum() < 1.0E-16))
        self.assertTrue(numpy.all(self.ilmenitePackageA.get_assay() - self.ilmenite.assays["IlmeniteA"] / self.ilmenite.assays["IlmeniteA"].sum() > -1.0E-16))

    def test_get_compound_mass(self):
        assay = "IlmeniteA"
        compound = "TiO2[S1]"
        for compound in self.ilmenite.compounds:
            mass = 1234.5 * self.ilmenite.assays[assay][self.ilmenite.get_compound_index(compound)] / self.ilmenite.get_assay_total(assay)
            self.assertEqual(self.ilmenitePackageA.get_compound_mass(compound), mass)

    def test_set_H(self):
        tempPackageA = self.ilmenitePackageA.clone()
        H = tempPackageA.H + 5.67
        tempPackageA.H = H

        self.assertEqual(tempPackageA.H, H)
        self.assertEqual(tempPackageA.T, 121.65612974840057)

    def test_get_H(self):
        self.assertEqual(self.ilmenitePackageA.H, self.ilmenitePackageA._H)

    def test_set_T(self):
        tempPackageA = self.ilmenitePackageA.clone()
        T = tempPackageA.T + 123.4
        tempPackageA.T = T

        self.assertEqual(tempPackageA.T, T)
        self.assertEqual(tempPackageA.H, -2778.3170030281622)

    def test_get_T(self):
        self.assertEqual(self.ilmenitePackageA.T, self.ilmenitePackageA._T)

    def test_get_P(self):
        self.assertEqual(self.ilmenitePackageA.P, self.ilmenitePackageA._P)

    def test_get_element_masses(self):
        x = self.ilmenitePackageA.get_element_masses()
        print(self.ilmenitePackageA.material.elements)
        print(x)
        y = self.ilmenitePackageA.get_element_mass("Ti")
        print(y)


# =============================================================================
# Display documentation and run tests.
# =============================================================================
# os.system("cls")

#help(Material)
#help(MaterialPackage)

if __name__ == '__main__':
    unittest.main()
